--[[
Name: LibSpreadsheet-1.0
Revision: $Rev: 3 $
Author(s): Humbedooh
Description: Spreadsheet stuff
Dependencies: LibStub AceGUI-3.0
License: GPL v3 or later.
]]
-----------------------------------------------------------------------

local SPREADSHEET = "LibSpreadsheet-1.0"
local SPREADSHEET_MINOR = tonumber(("$Rev: 3 $"):match("(%d+)"))
if not LibStub then error(SPREADSHEET .. " requires LibStub.") end
local lag = LibStub("AceGUI-3.0", true)
if not lag then error(SPREADSHEET .. " requires AceGUI-3.0.") end
local lib = LibStub:NewLibrary(SPREADSHEET, SPREADSHEET_MINOR)
if not lib then return end

lib.disabled = lib.disabled or nil;
lib.objects = lib.objects or {};
lib.sheets = lib.sheets or {};
lib.str = UIParent:CreateFontString();
lib.cols = {'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q','R','S','T','U','V','W','X','Y','Z', '?'};

function lib:AddSheet(title)
	local obj = self;
	if ( type(title) ~= "string" ) then error(("%s: Error in setting sheet title; Expected string, got %s!"):format(SPREADSHEET, type(title))); end
	for k, v in pairs(obj.sheets) do
		if ( v.title == title ) then error(("%s: Sheet \"%s\" already exists in book \"%s\"!"):format(SPREADSHEET, title, obj.title)); end
	end
	local sheet = {
        parent = obj,
		type = 'sheet',
		title = title,
		columns = {},
        lastColumn = 0,
		rows = {},
		rowHeight = 0,
		lastRow = 0,
		lastColumn = 0,
        currentRow = nil,
        currentColumn = nil,
		frame = lag:Create("SimpleGroup"),
		AddColumn = lib.AddColumn,
		SetColumn = lib.SetColumn,
		DeleteColumn = lib.DeleteColumn,	
        ExtractLocation = lib.ExtractLocation,
		AddRow = lib.AddRow,
		GoTo = lib.GoTo,
		Insert = lib.Insert,
		Delete = lib.Delete,
        isInteractive = false,
        StartSelect = lib.StartSelect,
        EndSelect = lib.EndSelect,
        SetVisibleSelection = lib.SetVisibleSelection,
        UpdateSelect = lib.UpdateSelect,
	}
	sheet.frame.width = "fill";
	sheet.frame.height = "fill";
--	sheet.scrollFrame.content.height = "fill";
--	sheet.scrollFrame:AddChild(sheet.innerFrame);
	table.insert(obj.sheets, sheet);
	return obj:Select(sheet.title);
end

function lib:DeleteSheet(title)
	local obj = self;
	if ( type(title) ~= "string" ) then error(("%s: Error in deleting sheet; Invalid sheet name - expected string, got %s!"):format(SPREADSHEET, type(title))); end
	for k, v in pairs(obj.sheets) do
		if ( v.title == title ) then
            if ( obj.parent.selectedSheet == v ) then
                obj.selectedSheet = nil;
            end
			table.remove(obj.sheets, k);
			break;
		end		
	end
end

function lib:Select(sheet)
	local obj = self;
	for k, v in pairs(obj.sheets) do
		if ( v.title == sheet ) then
            obj.selectedSheet = v;
			return v;
		end
	end
	return nil;
end

function lib:AddColumn(...)
    local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error while adding column, expected sheet, got %s!"):format(SPREADSHEET, t)); end
	for i = 1, select("#", ...) do
		local column = select(i, ...);
        if ( type(column) == "string" ) then
            lib.str:SetFontObject("LibSpreadsheet_StringFont");
            lib.str:SetText(column);
            local w = lib.str:GetStringWidth() + 5;
			local h = lib.str:GetStringHeight() + 2;
            if ( h > obj.rowHeight ) then obj.rowHeight = h; end
            table.insert(obj.columns, {type="string", width=w, name=column});
            obj.lastColumn = obj.lastColumn + 1;
        elseif ( type(column) == "table" ) then
            local col = {name="", type="string"};
            if ( column.name ~= nil ) then col.name = column.name; end
            if ( column.type == "string" or column.type == "number" or column.type == "boolean" ) then
                col.type = column.type;
            end
            if ( col.type == "number" ) then
                lib.str:SetFontObject("LibSpreadsheet_NumberFont");
            else
                lib.str:SetFontObject("LibSpreadsheet_StringFont");
            end
            col.suffix = column.suffix or "";
            lib.str:SetText(col.name .. col.suffix);
            local w = lib.str:GetStringWidth() + 5;
			local h = lib.str:GetStringHeight() + 2;
            if ( h > obj.rowHeight ) then obj.rowHeight = h; end
            col.width = w;
            if ( column.decimals ~= nil ) then
                col.decimals = column.decimals;
            end
            obj.lastColumn = obj.lastColumn + 1;
            table.insert(obj.columns, col);
        else
            error(("%s: Error while adding column, expected string or table, got %s!"):format(SPREADSHEET, column));
        end
		
	end
end

function lib:SetColumn(column, properties)
	local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error while setting column, expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = obj;
    local c = nil;
    if ( type(column) == "string" ) then
        c = lib.cols[string.upper(column)];
    else
        c = tonumber(column);
        if ( c > 26 ) then c = 26; end
    end
    if ( sheet.columns[c] == nil ) then error(("%s: Error while setting column %s: No such column!"):format(SPREADSHEET, lib.cols[c])); end
    if ( type(properties) == "string" ) then
        lib.str:SetText(properties);
        local w = lib.str:GetStringWidth() + 5;
        local h = lib.str:GetStringHeight() + 2;
        if ( h > sheet.rowHeight ) then sheet.rowHeight = h; end
        sheet.columns[c].width = w;
        sheet.columns[c].name=properties;
    elseif ( type(column) == "table" ) then
        if ( column.name ~= nil ) then sheet.columns[c].name = column.name; end
        if ( column.type == "string" or column.type == "number" or column.type == "boolean" ) then
            sheet.columns[c].type = column.type;
        end
        if ( sheet.columns[c].type == "number" ) then
            lib.str:SetFontObject("LibSpreadsheet_NumberFont");
        else
            lib.str:SetFontObject("LibSpreadsheet_StringFont");
        end
        sheet.columns[c].suffix = column.suffix or sheet.columns[c].suffix;
        lib.str:SetText(sheet.columns[c].name .. sheet.columns[c].suffix);
        local w = lib.str:GetStringWidth() + 5;
        local h = lib.str:GetStringHeight() + 2;
        if ( h > sheet.rowHeight ) then sheet.rowHeight = h; end
        sheet.columns[c].width = w;
        if ( column.decimals ~= nil ) then
            sheet.columns[c].decimals = column.decimals;
        end
    else
        error(("%s: Error while setting column properties, expected string or table, got %s!"):format(SPREADSHEET, properties));
    end
end

function deleteColumn(column)
	local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error while deleting column, expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = obj;
    local c = nil;
    if ( type(column) == "string" ) then
        c = lib.cols[string.upper(column)];
    else
        c = tonumber(column);
        if ( c > 26 ) then c = 26; end
    end
    if ( sheet.columns[c] == nil ) then error(("%s: Error while deleting column %s: No such column!"):format(SPREADSHEET, lib.cols[c])); end
    table.remove(sheet.columns, c);
    for k, v in pairs(sheet.rows) do
        table.remove(v, c);
    end
end

function lib:AddRow(...)
	local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error while adding row, expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = obj;
    local row = {};
    for i = 1, select("#", ...) do
		local cell = select(i, ...);
        if ( sheet.parent.strict == true ) then
            local t =  sheet.columns[i].type;
            if ( t == "string" and type(cell) ~= "string" ) then error(("%s: Error while adding cell in column %s, expected string, got %s!"):format(SPREADSHEET, lib.cols[i], type(cell))); end
            if ( t == "boolean" and type(cell) ~= "boolean" ) then error(("%s: Error while adding cell in column %s, expected boolean, got %s!"):format(SPREADSHEET, lib.cols[i], type(cell))); end
            if ( ( t == "number" ) and type(cell)  ~= "number" ) then error(("%s: Error while adding cell in column %s, expected number, got %s!"):format(SPREADSHEET, lib.cols[i], type(cell))); end
        end
        if ( sheet.columns[i] ~= nil ) then
            if ( sheet.columns[i].type == "string" ) then
                lib.str:SetFontObject("LibSpreadsheet_StringFont");
                if ( type(cell) == "boolean" ) then
                    if ( cell == true ) then cell = "true";
                    else cell = "false";
                    end
                end
                lib.str:SetText(cell);
                local w = lib.str:GetStringWidth() + 5;
                if ( w > sheet.columns[i].width ) then sheet.columns[i].width= w; end
            elseif ( sheet.columns[i].type == "number" ) then
                lib.str:SetFontObject("LibSpreadsheet_NumberFont");
                lib.str:SetText(cell);
                local w = lib.str:GetStringWidth() + 5;
                if ( w > sheet.columns[i].width ) then sheet.columns[i].width= w; end
            end
        end
        table.insert(row, {data=cell});
    end
    table.insert(sheet.rows, row);
end

function lib:ExtractLocation(str)
    local t = (self.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error while looking up column, expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = self;
    local col = nil;
    local row = nil;
    if ( type(str) == "string" ) then
        local c, r = (str):match("(%D+)(%d*)");
        if ( r ~= nil ) then
            row = r;
        end
        if ( string.len(c) == 1 ) then
            col = tonumber(c, 36) - 9;
        else
            for k, column in pairs(sheet.columns) do
                if ( column.name == c ) then
                    col = k;
                    break;
                end
            end
        end
    elseif ( type(column) == "number" ) then
        if ( column > 26 ) then column = math.fmod(column, 26); end
        if ( column < 1 ) then 
            while ( column < 1 ) do
                column = column + sheet.lastColumn;
            end
        end
    end
    return col, tonumber(row);
end

function lib:SetVisibleSelection(fromCol, fromRow, toCol, toRow)
    local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error in GoTo(): Expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = obj;
    if ( type(fromCol) == "number" ) then
        sheet.selectStartColumn = fromCol;
        sheet.selectStartRow = fromRow;
    elseif ( type(fromCol) == "string" ) then
        local c, r = sheet:ExtractLocation(fromCol);
        if ( type(r) == "number" ) then
            sheet.selectStartColumn = c;
            sheet.selectStartRow = r;
            toRow = toCol;
            toCol = fromRow;
        else
            sheet.selectStartColumn = c;
            sheet.selectStartRow = fromRow;
        end
    end
    if ( type(toCol) == "number" ) then
        sheet.selectEndColumn = toCol;
        sheet.selectEndRow = toRow;
    elseif ( type(toCol) == "string" ) then
        local c, r = sheet:ExtractLocation(toCol);
        if ( type(r) == "number" ) then
            sheet.selectEndColumn = c;
            sheet.selectEndRow = r;
        else
            sheet.selectEndColumn = c;
            sheet.selectEndRow = toRow;
        end
    end
    sheet.selecting = true;
    sheet:UpdateSelect(nil);
    sheet.selecting = false;
end

function lib:GoTo(column, row) -- takes arguments like: GoTo("B"), GoTo("B1"), GoTo("B", 3), GoTo(2,5), GoTo("Name", 3), GoTo("Level5");
    local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error in GoTo(): Expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = obj;
	local c, r = sheet:ExtractLocation(column);
    if ( type(r) == "number" ) then
        row = r;
    end
    if ( sheet.columns[column] == nil ) then error(("%s: Error in GoTo(): No such column '%s'!"):format(SPREADSHEET, lib.cols[column or 27])); end
    if ( sheet.rows[row] == nil ) then error(("%s: Error in GoTo(): No such row '%u'!"):format(SPREADSHEET, row)); end
    sheet.currentColumn = column;
    sheet.currentRow = row;    
end

function lib:Delete(fromCol, fromRow, toCol, toRow)
	
end

function lib:Insert(...)
	
end

function lib:StartSelect(caller)
    local sheet = self;
    sheet.selecting = true;
    sheet.selectStartColumn = caller.column;
    sheet.selectStartRow = caller.row;
    sheet:UpdateSelect(caller);
end

function lib:EndSelect(caller)
    local sheet = self;
    sheet.selectEndColumn = caller.column;
    sheet.selectEndRow = caller.row;
    sheet.selecting = false;
    sheet:UpdateSelect(caller);
end

function lib:UpdateSelect(caller)
    local sheet = self;
    if ( sheet.selecting or false == true ) then
        if ( caller ~= nil ) then
            sheet.selectEndColumn = caller.column;
            sheet.selectEndRow = caller.row;
        end
        for x = 1, table.maxn(sheet.rows) do
            for y = 1, sheet.lastColumn do
                sheet.rows[x][y].button.tex:SetTexture(0,0,0,0);
            end
        end
        if ( sheet.selectEndRow == nil ) then sheet.selectEndRow  = sheet.selectStartRow ; end
        for x = math.min(sheet.selectStartRow, sheet.selectEndRow), math.max(sheet.selectStartRow, sheet.selectEndRow)do
            for y = math.min(sheet.selectStartColumn, sheet.selectEndColumn), math.max(sheet.selectStartColumn, sheet.selectEndColumn) do
                sheet.rows[x][y].button.tex:SetTexture(0.3,0.2,1,0.25);
            end
        end
    end
end


function lib:Build()
    local obj = self;
    local sheet = obj.selectedSheet;
    sheet.frame:ReleaseChildren();
    obj.innerFrame:ReleaseChildren();
    local header = lag:Create("SimpleGroup");
    header:SetLayout("Flow");
    header.width = "fill";
    for key, cell in pairs (sheet.columns) do
        local label = nil;
        if ( cell.click ~= nil ) then
            label = lag:Create("InteractiveLabel");
            label:SetCallback("OnClick", cell.click);
        else
            label = lag:Create("Label");
        end
        label:SetFontObject(LibSpreadsheet_StringFont);
        label:SetWidth(cell.width);
        label:SetText("|cffEEAA22" ..(cell.name or "") .. ":|r");
        if ( cell.type == "number" or cell.type == "float" ) then label.label:SetJustifyH("RIGHT"); end
        if ( cell.type == "string" ) then label.label:SetJustifyH("LEFT"); end
        header:AddChild(label);
        local spacing = lag:Create("Label");
        spacing:SetText("");
        spacing:SetFontObject(LibSpreadsheet_StringFont);
        spacing:SetWidth(sheet.rowHeight);
        header:AddChild(spacing);
    end
    obj.frame:AddChild(header, obj.scrollFrame);
    
    for rkey, rowitem in pairs (sheet.rows) do
        local row = lag:Create("SimpleGroup");
        row:SetLayout("Flow");
        row.width = "fill";
        local i = 0;
        for key, cell in pairs(rowitem) do
            i = i + 1;
            local label = lag:Create("InteractiveLabel");
            if ( cell.click ~= nil ) then
                label:SetCallback("OnClick", cell.click);
            else
                label.tex = label.frame:CreateTexture(nil, "OVERLAY");
                label.tex:SetAllPoints();
                label.tex:SetBlendMode("ADD");
                label.column = i;
                label.row = rkey;
                if ( sheet.isInteractive or sheet.columns[i].interactive ) then
                    label.frame:SetScript("OnMouseDown", function(caller, event) sheet:StartSelect(label); end);
                    label.frame:SetScript("OnMouseUp", function(caller, event) sheet:EndSelect(label); end);
                    label.frame:SetScript("OnEnter", function(caller, event) sheet:UpdateSelect(label); end);
                    label.frame:SetScript("OnLeave", function(caller, event) sheet:UpdateSelect(label); end);
                end
                cell.button = label;
            end
            label:SetFontObject(LibSpreadsheet_StringFont);
            label:SetWidth(sheet.columns[i].width);
            
            label:SetHeight(sheet.rowHeight);
            if ( sheet.columns[i].type == "number" ) then
                label.label:SetJustifyH("RIGHT");
                label:SetFontObject(LibSpreadsheet_NumberFont);
                if ( sheet.columns[i].decimals ~= nil ) then
                    local s = ("%%.%uf%%s"):format(sheet.columns[i].decimals);
                    label:SetText(string.format(s, cell.data, (sheet.columns[i].suffix or "")));
                else
                    label:SetText(string.format("%i%s", cell.data, (sheet.columns[i].suffix or "")));
                end
            elseif ( sheet.columns[i].type == "string" ) then
                label.label:SetJustifyH("LEFT");
                label:SetText(cell.data);
            elseif ( sheet.columns[i].type == "boolean" ) then
                label.label:SetJustifyH("LEFT");
                if ( cell.data == true ) then label:SetText("true");
                else label:SetText("false"); end
            end
            row:AddChild(label);
            local spacing = lag:Create("Label");
            spacing:SetText("");
            spacing:SetFontObject(LibSpreadsheet_StringFont);
            spacing:SetWidth(sheet.rowHeight);
            row:AddChild(spacing);
        end
        row:SetHeight(sheet.rowHeight);
        row:SetAutoAdjustHeight(false);
        sheet.frame:AddChild(row);
    end
    --obj.scrollFrame.content.height="fill";
    obj.innerFrame:AddChild(sheet.frame);
    obj.scrollFrame:SetHeight(obj.frame.frame:GetHeight() - 20);
    return obj.frame;
end

function lib:StrictMode(toggle)
	if ( type(toggle) ~= "boolean" ) then error(("%s: Error in setting strict mode; Expected boolean, got %s!"):format(SPREADSHEET, type(toggle))); end
	self.strict = toggle;
end

function lib:Book(title)
	local obj = {};
	if ( type(title) ~= "string" ) then error(("%s: Error in creating book title; Expected string, got %s!"):format(SPREADSHEET, type(title))); end
	obj.sheets = {};
    obj.selectedSheet = nil;
	obj.type = 'book';
	obj.title = title;
    obj.strict = false;
	obj.Add = lib.AddSheet;
	obj.Delete = lib.DeleteSheet;
	obj.Select = lib.Select;
	obj.StrictMode = lib.StrictMode;
    obj.Build = lib.Build;
    obj.frame = lag:Create("SimpleGroup");
    obj.frame.height = "fill";
    obj.frame.width = "fill";
    obj.frame:SetLayout("Flow");
    obj.scrollFrame = lag:Create("ScrollFrame");
    obj.scrollFrame.height = "fill";
    obj.scrollFrame.width = "fill";
	obj.innerFrame = lag:Create("SimpleGroup");
    obj.innerFrame.height = "fill";
    obj.innerFrame.width = "fill";
    obj.scrollFrame:SetLayout("Fill");
    obj.scrollFrame:AddChild(obj.innerFrame);
    obj.frame:AddChild(obj.scrollFrame);
	return obj;
end

function LibSpreadTest()
    local lss = LibStub("LibSpreadsheet-1.0", true);                    -- Load the lib
    local book = lss:Book("myBook");                                    -- Make a new book
    local sheet = book:Add("my stuff");                                 -- Add a sheet to the book
    book:StrictMode(true);
    sheet.isInteractive = true;
    sheet:AddColumn(
        "Name",
        "Class",
        {name="Level", type="number"},
        {name="Score", type="number", decimals=2, suffix="%"},
        {name="Uses LibSpreadsheet!", type="boolean"}
        );  -- Set some headers
    sheet:AddRow("Humbedooh", "Warrior", 80, 1.34, true);                     -- Add some data
    sheet:AddRow("Kitiania", "|cffFF4499Priest|r", 80, 82.5, false);                      -- Add another row
    sheet:AddRow("Wimzig", "Gnome....", 1, 21, true);                     -- Add some data
    sheet:AddRow("Humbedooh", "Warrior", 80, 1.34, true);                     -- Add some data
    sheet:AddRow("Kitiania", "|cffFF4499Priest|r", 80, 82.5, false);                      -- Add another row
    sheet:AddRow("Wimzig", "Gnome....", 1, 21, true);                     -- Add some data
    sheet:AddRow("Humbedooh", "Warrior", 80, 1.34, true);                     -- Add some data
    sheet:AddRow("Kitiania", "|cffFF4499Priest|r", 80, 82.5, false);                      -- Add another row
    sheet:AddRow("Wimzig", "Gnome....", 1, 21, true);                     -- Add some data
    
    
    local frame = book:Build();                                        -- Create the UI
    local myWindow = lag:Create("Frame");
    myWindow:SetWidth(500);
    myWindow:SetHeight(300);
    myWindow:SetLayout("fill");
    myWindow:AddChild(frame);
    myWindow:Show();
    sheet:SetVisibleSelection("A2", "C6");
end